import time
import argparse

import torch
from tqdm import tqdm
import datetime

import torch.nn as nn
import torch.backends.cudnn as cudnn
from torch.utils.data import DataLoader
from Utils.tools import data_write_csv
from loss import *
from utils import *
from biomass_dataset import BiomassDB
from CONR_train.biomass_dir.models.resnet import resnet50 as create_model
from ranksim import batchwise_ranking_regularizer
import os
from torchmetrics.regression import MeanAbsolutePercentageError


parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
# imbalanced related


# LDS
parser.add_argument('--lds', action='store_true', default=False, help='whether to enable LDS')
parser.add_argument('--lds_kernel', type=str, default='gaussian',
                    choices=['gaussian', 'triang', 'laplace'], help='LDS kernel type')
parser.add_argument('--lds_ks', type=int, default=9, help='LDS kernel size: should be odd number')
parser.add_argument('--lds_sigma', type=float, default=1, help='LDS gaussian/laplace kernel sigma')

# FDS
parser.add_argument('--fds', action='store_true', default=False, help='whether to enable FDS')
parser.add_argument('--fds_kernel', type=str, default='gaussian',
                    choices=['gaussian', 'triang', 'laplace'], help='FDS kernel type')
parser.add_argument('--fds_ks', type=int, default=9, help='FDS kernel size: should be odd number')
parser.add_argument('--fds_sigma', type=float, default=1, help='FDS gaussian/laplace kernel sigma')
parser.add_argument('--start_update', type=int, default=0, help='which epoch to start FDS updating')
parser.add_argument('--start_smooth', type=int, default=1, help='which epoch to start using FDS to smooth features')
parser.add_argument('--bucket_num', type=int, default=100, help='maximum bucket considered for FDS')
parser.add_argument('--bucket_start', type=int, default=3, choices=[0, 3],
                    help='minimum(starting) bucket for FDS, 0 for IMDBWIKI, 3 for AgeDB')
parser.add_argument('--fds_mmt', type=float, default=0.9, help='FDS momentum')

# re-weighting: SQRT_INV / INV
parser.add_argument('--reweight', type=str, default='sqrt_inv', choices=['none', 'sqrt_inv', 'inverse'],
                    help='cost-sensitive reweighting scheme')
# two-stage training: RRT
parser.add_argument('--retrain_fc', action='store_true', default=False,
                    help='whether to retrain last regression layer (regressor)')

# RankSim
parser.add_argument('--regularization_weight', type=float, default=0, help='weight of the regularization term')
parser.add_argument('--interpolation_lambda', type=float, default=1.0, help='interpolation strength')

# training/optimization related
parser.add_argument('--dataset', type=str, default='pepper', choices=['pepper', 'grass'], help='dataset name')
parser.add_argument('--data_dir', type=str, default='./data', help='data directory')
parser.add_argument('--model', type=str, default='resnet', help='model name')
parser.add_argument('--store_root', type=str, default='checkpoint', help='root path for storing checkpoints, logs')
parser.add_argument('--store_name', type=str, default='', help='experiment store name')
parser.add_argument('--gpu', type=int, default=None)
parser.add_argument('--optimizer', type=str, default='Adam', choices=['Adam', 'SGD'], help='optimizer type')
parser.add_argument('--loss', type=str, default='focal_l1', choices=['mse', 'l1', 'focal_l1', 'focal_mse', 'huber'],
                    help='training loss type')
parser.add_argument('--lr', type=float, default=1e-3, help='initial learning rate')
parser.add_argument('--epoch', type=int, default=60, help='number of epochs to train')
parser.add_argument('--momentum', type=float, default=0.9, help='optimizer momentum')
parser.add_argument('--weight_decay', type=float, default=1e-4, help='optimizer weight decay')
parser.add_argument('--schedule', type=int, nargs='*', default=[60, 80], help='lr schedule (when to drop lr by 10x)')
parser.add_argument('--batch_size', type=int, default=32, help='batch size')
parser.add_argument('--print_freq', type=int, default=10, help='logging frequency')
parser.add_argument('--img_size', type=int, default=224, help='image size used in training')
parser.add_argument('--workers', type=int, default=1.5, help='number of workers used in data loading')

# checkpoints
parser.add_argument('--resume', type=str, default='', help='checkpoint file path to resume training')
parser.add_argument('--pretrained', type=str, default='', help='checkpoint file path to load backbone weights')
parser.add_argument('--evaluate', action='store_true', help='evaluate only flag')
parser.add_argument('--IsSaveCkpt', action='store_true', default=False, help='Is Save Checkpoint?')

# ConR
parser.add_argument('--conr', action='store_true', default=True, help='whether to enable conr')
parser.add_argument('-w', type=float, default=1, help='similarity window for conR loss')
parser.add_argument('--beta', type=float, default=4, help='conR loss coeff')
parser.add_argument('-t', type=float, default=.2, help='temperature')
parser.add_argument('-e', type=float, default=0.01, help="coeff for eta in ConR")

parser.set_defaults(augment=True)
args, unknown = parser.parse_known_args()

args.start_epoch, args.best_loss = 0, 1e5

if len(args.store_name):
    args.store_name = f'_{args.store_name}'
if not args.lds and args.reweight != 'none':
    args.store_name += f'_{args.reweight}'
if args.conr:
    args.store_name += f'ConR_{args.beta}_w={args.w}'
if args.lds:
    args.store_name += f'_lds_{args.lds_kernel[:3]}_{args.lds_ks}'
    if args.lds_kernel in ['gaussian', 'laplace']:
        args.store_name += f'_{args.lds_sigma}'
if args.fds:
    args.store_name += f'_fds_{args.fds_kernel[:3]}_{args.fds_ks}'
    if args.fds_kernel in ['gaussian', 'laplace']:
        args.store_name += f'_{args.fds_sigma}'
    args.store_name += f'_{args.start_update}_{args.start_smooth}_{args.fds_mmt}'
if args.retrain_fc:
    args.store_name += f'_retrain_fc'
if args.regularization_weight > 0:
    args.store_name += f'_reg{args.regularization_weight}_il{args.interpolation_lambda}'
args.store_name = f"{args.dataset}_{args.model}{args.store_name}_{args.optimizer}_{args.loss}_{args.lr}_{args.batch_size}"

timestamp = str(datetime.datetime.now())
timestamp = '-'.join(timestamp.split(' '))
args.store_name = args.store_name + '_' + timestamp

prepare_folders(args)

logging.root.handlers = []
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s | %(message)s",
    handlers=[
        logging.FileHandler(os.path.join(args.store_root, args.store_name, 'training.log')),
        logging.StreamHandler()
    ])
print = logging.info
print(f"Args: {args}")
print(f"Store name: {args.store_name}")


def main():
    if args.gpu is not None:
        print(f"Use GPU: {args.gpu} for training")

    # Data
    # pepper

    train_root_dir = '/Dataset/mix_pepper/train'
    test_root_dir = '/Dataset/mix_pepper/test'
    train_csv = '/Dataset/mix_pepper/train_pepper.csv'
    test_csv = '/Dataset/mix_pepper/test_pepper.csv'

    print('=====> Preparing data...')

    df_train = pd.read_csv(train_csv)
    df_val = pd.read_csv(test_csv)
    df_test = pd.read_csv(test_csv)
    train_labels = df_train['Biomass']

    train_dataset = BiomassDB(data_dir=train_root_dir, df=df_train, img_size=args.img_size, split='train',
                              reweight=args.reweight, lds=args.lds, lds_kernel=args.lds_kernel, lds_ks=args.lds_ks,
                              lds_sigma=args.lds_sigma)
    val_dataset = BiomassDB(data_dir=test_root_dir, df=df_val, img_size=args.img_size, split='val')
    test_dataset = BiomassDB(data_dir=test_root_dir, df=df_test, img_size=args.img_size, split='val')

    train_loader = DataLoader(train_dataset, batch_size=args.batch_size, shuffle=True,
                              num_workers=args.workers, pin_memory=True, drop_last=False)
    val_loader = DataLoader(val_dataset, batch_size=args.batch_size, shuffle=False,
                            num_workers=args.workers, pin_memory=True, drop_last=False)
    test_loader = DataLoader(test_dataset, batch_size=args.batch_size, shuffle=False,
                             num_workers=args.workers, pin_memory=True, drop_last=False)
    print(f"Training data size: {len(train_dataset)}")
    print(f"Validation data size: {len(val_dataset)}")
    print(f"Test data size: {len(test_dataset)}")

    # Model
    print('=====> Building model...')

    model = create_model(num_classes=1)
    model = model.cuda()
    count = 0

    # evaluate only
    if args.evaluate:
        assert args.resume, 'Specify a trained model using [args.resume]'
        checkpoint = torch.load(args.resume)
        model.load_state_dict(checkpoint['state_dict'], strict=False)
        print(f"===> Checkpoint '{args.resume}' loaded (epoch [{checkpoint['epoch']}]), testing...")
        validate(test_loader, model, train_labels=train_labels, prefix='Test')
        return

    if args.retrain_fc:
        assert args.reweight != 'none' and args.pretrained
        print('===> Retrain last regression layer only!')
        for name, param in model.named_parameters():
            if 'fc' not in name and 'linear' not in name:
                param.requires_grad = False

    # Loss and optimizer
    if not args.retrain_fc:
        optimizer = torch.optim.Adam(model.parameters(), lr=args.lr) if args.optimizer == 'adam' else \
            torch.optim.SGD(model.parameters(), lr=args.lr, momentum=args.momentum, weight_decay=args.weight_decay)
    else:
        # optimize only the last linear layer
        parameters = list(filter(lambda p: p.requires_grad, model.parameters()))
        names = list(filter(lambda k: k is not None,
                            [k if v.requires_grad else None for k, v in model.module.named_parameters()]))
        assert 1 <= len(parameters) <= 2  # fc.weight, fc.bias
        print(f'===> Only optimize parameters: {names}')
        optimizer = torch.optim.Adam(parameters, lr=args.lr) if args.optimizer == 'adam' else \
            torch.optim.SGD(parameters, lr=args.lr, momentum=args.momentum, weight_decay=args.weight_decay)

    if args.pretrained:
        checkpoint = torch.load(args.pretrained, map_location="cpu")
        from collections import OrderedDict
        new_state_dict = OrderedDict()
        for k, v in checkpoint['state_dict'].items():
            if 'linear' not in k and 'fc' not in k:
                new_state_dict[k] = v
        model.load_state_dict(new_state_dict, strict=False)
        print(f'===> Pretrained weights found in total: [{len(list(new_state_dict.keys()))}]')
        print(f'===> Pre-trained model loaded: {args.pretrained}')

    if args.resume:
        if os.path.isfile(args.resume):
            print(f"===> Loading checkpoint '{args.resume}'")
            checkpoint = torch.load(args.resume) if args.gpu is None else \
                torch.load(args.resume, map_location=torch.device(f'cuda:{str(args.gpu)}'))
            args.start_epoch = checkpoint['epoch']
            args.best_loss = checkpoint['best_loss']
            model.load_state_dict(checkpoint['state_dict'])
            optimizer.load_state_dict(checkpoint['optimizer'])
            print(f"===> Loaded checkpoint '{args.resume}' (Epoch [{checkpoint['epoch']}])")
        else:
            print(f"===> No checkpoint found at '{args.resume}'")

    cudnn.benchmark = True

    best_mape = 1e6
    best_rmse = 1e6
    scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer=optimizer, T_max=args.epoch, eta_min=1e-8)

    val_avg_rmse_Data = [[]]
    val_avg_mae_Data = [[]]
    val_avg_mape_Data = [[]]


    for epoch in range(args.start_epoch, args.epoch):
        adjust_learning_rate(optimizer, epoch, args)
        train_loss = train(train_loader, model, optimizer, epoch, train_labels)
        val_loss_mse, val_loss_l1, val_loss_gmean, val_mape, val_rmse = validate(val_loader, model,
                                                                                         train_labels=train_labels)

        scheduler.step()
        loss_metric = val_loss_mse if args.loss == 'mse' else val_loss_l1
        # loss_metric = low_loss
        is_best = loss_metric < args.best_loss
        args.best_loss = min(loss_metric, args.best_loss)
        best_mape = min(val_mape, best_mape)
        best_rmse = min(val_rmse, best_rmse)

        print(
            f"Best {'L1' if 'l1' in args.loss else 'MSE'} Loss: {args.best_loss: .3f} MAPE: {best_mape: .3f} RMSE: {best_rmse:.3f}")

        if args.IsSaveCkpt:
            save_checkpoint(args, {
                'epoch': epoch + 1,
                'model': args.model,
                'best_loss': args.best_loss,
                'state_dict': model.state_dict(),
                'optimizer': optimizer.state_dict(),
            }, is_best)

        print(f"Epoch #{epoch}: Train loss [{train_loss:.4f}]; "
              f"Val loss: RMSE [{val_rmse:.4f}], L1 [{val_loss_l1:.4f}], G-Mean [{val_loss_gmean:.4f}] MAPE [{val_mape:.4f}] ")

        val_avg_rmse_Data.append([epoch + 1, val_rmse])
        val_avg_mae_Data.append([epoch + 1, val_loss_l1])
        val_avg_mape_Data.append([epoch + 1, val_mape])


    crop_class = args.dataset
    opt = args.optimizer
    model_name = args.model
    Loss_name = args.loss

    save_csv_path = "/biomass_dir/run/" + crop_class + "/" + opt + "/" + model_name + f"/lr{args.lr}/"
    common_prefix = save_csv_path + Loss_name + f"_ConR_beta={args.beta}_w={args.w}_eta={args.e}_t={args.t}_epoch_{args.epoch}"
    os.makedirs(save_csv_path, exist_ok=True)

    val_rmse_path = common_prefix + "_val_rmse.csv"
    val_mae_path = common_prefix + "_val_mae.csv"
    val_mape_path = common_prefix + "_val_mape.csv"


    data_write_csv(val_rmse_path, val_avg_rmse_Data)
    data_write_csv(val_mae_path, val_avg_mae_Data)
    data_write_csv(val_mape_path, val_avg_mape_Data)



def train(train_loader, model, optimizer, epoch, train_labels):
    batch_time = AverageMeter('Time', ':6.2f')
    data_time = AverageMeter('Data', ':6.4f')
    losses = AverageMeter(f'Loss ({args.loss.upper()})', ':.3f')

    if args.conr:
        loss_c = AverageMeter("Loss(ConR)", ":.3f")
        progress = ProgressMeter(
            len(train_loader),
            [batch_time, data_time, losses, loss_c],
            prefix="Epoch: [{}]".format(epoch)
        )
    else:
        progress = ProgressMeter(
            len(train_loader),
            [batch_time, data_time, losses],
            prefix="Epoch: [{}]".format(epoch)
        )

    model.train()
    end = time.time()
    preds, labels = [], []

    for idx, (inputs, imgs_c, targets, weights) in enumerate(train_loader):

        data_time.update(time.time() - end)
        inputs, targets, weights = \
            inputs.cuda(non_blocking=True), targets.cuda(non_blocking=True), weights.cuda(non_blocking=True)

        imgs_c[0] = imgs_c[0].cuda(args.gpu, non_blocking=True)
        imgs_c[1] = imgs_c[1].cuda(args.gpu, non_blocking=True)
        imgs_c = torch.cat((imgs_c[0], imgs_c[1]), dim=0)
        targets_c = torch.cat((targets, targets), dim=0)
        weights_c = torch.cat((weights, weights), dim=0)


        outputs_c, features_c = model(imgs_c)
        outputs, features = model(inputs)

        loss = globals()[f"weighted_{args.loss}_loss"](outputs, targets, weights)

        if args.conr:
            l_c = SCDR(features_c, targets_c, outputs_c.clone().detach(), w=args.w, weights=weights_c, t=args.t,
                       e=args.e)
            loss += args.beta * l_c
            loss_c.update(l_c.item(), inputs[0].size(0))

        if args.regularization_weight > 0:
            loss += (args.regularization_weight * batchwise_ranking_regularizer(features, targets,
                                                                                args.interpolation_lambda))

        assert not (np.isnan(loss.item()) or loss.item() > 1e10), f"Loss explosion: {loss.item()}"

        losses.update(loss.item(), inputs.size(0))

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        batch_time.update(time.time() - end)
        end = time.time()
        if idx % args.print_freq == 0:
            progress.display(idx)

    if args.fds and epoch >= args.start_update:
        print(f"Create Epoch [{epoch}] features of all training data...")
        encodings, labels = [], []
        with torch.no_grad():
            for (inputs, _, targets, _) in tqdm(train_loader):
                inputs = inputs.cuda(non_blocking=True)
                outputs, feature = model(inputs, targets, epoch)
                encodings.extend(feature.data.squeeze().cpu().numpy())
                labels.extend(targets.data.squeeze().cpu().numpy())

        encodings, labels = torch.from_numpy(np.vstack(encodings)).cuda(), torch.from_numpy(np.hstack(labels)).cuda()
        model.module.FDS.update_last_epoch_stats(epoch)
        model.module.FDS.update_running_stats(encodings, labels, epoch)

    return losses.avg


def validate(val_loader, model, train_labels=None, prefix='Val'):
    batch_time = AverageMeter('Time', ':6.3f')
    losses_mse = AverageMeter('Loss (MSE)', ':.3f')
    losses_l1 = AverageMeter('Loss (L1)', ':.3f')
    losses_mape = AverageMeter('MAPE', ':.3f')
    losses_rmse = AverageMeter('Loss (RMSE)', ':.3f')
    progress = ProgressMeter(
        len(val_loader),
        [batch_time, losses_mse, losses_l1],
        prefix=f'{prefix}: '
    )

    criterion_mse = nn.MSELoss()
    criterion_l1 = nn.L1Loss()
    criterion_gmean = nn.L1Loss(reduction='none')
    criterion_mape = MeanAbsolutePercentageError().cuda()


    model.eval()
    losses_all = []
    preds, labels = [], []
    f = []

    with torch.no_grad():
        end = time.time()
        pred_dict = {}
        for idx, (inputs, targets) in enumerate(val_loader):
            inputs, targets = inputs.cuda(non_blocking=True), targets.cuda(non_blocking=True)

            if args.evaluate:
                outputs, features = model(inputs)
                f.extend(features.data.cpu().numpy())

            else:
                outputs, _ = model(inputs)

            preds.extend(outputs.data.cpu().numpy())
            labels.extend(targets.data.cpu().numpy())

            loss_mse = criterion_mse(outputs, targets)
            loss_l1 = criterion_l1(outputs, targets)
            loss_all = criterion_gmean(outputs, targets)
            mape = criterion_mape(outputs, targets)
            rmse = torch.sqrt(loss_mse)
            losses_all.extend(loss_all.cpu().numpy())

            losses_mse.update(loss_mse.item(), inputs.size(0))
            losses_l1.update(loss_l1.item(), inputs.size(0))
            losses_mape.update(mape.item(), inputs.size(0))
            losses_rmse.update(rmse.item(), inputs.size(0))
            batch_time.update(time.time() - end)
            end = time.time()
            if idx % args.print_freq == 0:
                progress.display(idx)

        loss_gmean = gmean(np.hstack(losses_all), axis=None).astype(float)
        print(
            f" * Val Overall: MSE {losses_mse.avg:.3f}\tL1 {losses_l1.avg:.3f}\tG-Mean {loss_gmean:.3f} \tMAPE {losses_mape.avg:.3f} ")

    return losses_mse.avg, losses_l1.avg, loss_gmean, losses_mape.avg, losses_rmse.avg


if __name__ == '__main__':
    main()
