import torch
import torch.nn as nn
from einops.layers.torch import Rearrange, Reduce
from timm.models.vision_transformer import trunc_normal_
from models.pytorch_image_models.timm.layers.drop import DropPath, DropBlock2d
import warnings


warnings.filterwarnings("ignore")


class deconv2d_bn(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size=2, strides=2):
        super(deconv2d_bn, self).__init__()
        self.conv1 = nn.ConvTranspose2d(in_channels, out_channels,
                                        kernel_size=kernel_size,
                                        stride=strides, bias=True)
        self.bn1 = nn.BatchNorm2d(out_channels)
        self.act = nn.LeakyReLU()

    def forward(self, x):
        out = self.act(self.bn1(self.conv1(x)))
        return out


class simam_module(torch.nn.Module):
    def __init__(self, channels=None, e_lambda=1e-4):
        super(simam_module, self).__init__()

        self.activaton = nn.Sigmoid()
        self.e_lambda = e_lambda

    def __repr__(self):
        s = self.__class__.__name__ + '('
        s += ('lambda=%f)' % self.e_lambda)
        return s

    @staticmethod
    def get_module_name():
        return "simam"

    def forward(self, x):
        b, c, h, w = x.size()

        n = w * h - 1

        x_minus_mu_square = (x - x.mean(dim=[2, 3], keepdim=True)).pow(2)
        y = x_minus_mu_square / (4 * (x_minus_mu_square.sum(dim=[2, 3], keepdim=True) / n + self.e_lambda)) + 0.5

        return x * self.activaton(y)


class Conv2d_BN(torch.nn.Sequential):
    def __init__(self, a, b, ks=1, stride=1, pad=0, dilation=1,
                 groups=1, bn_weight_init=1):
        super().__init__()
        self.add_module('c', torch.nn.Conv2d(
            a, b, ks, stride, pad, dilation, groups, bias=False))
        self.add_module('bn', torch.nn.BatchNorm2d(b))

        # new_added

        torch.nn.init.constant_(self.bn.weight, bn_weight_init)
        torch.nn.init.constant_(self.bn.bias, 0)

    @torch.no_grad()
    def fuse(self):
        c, bn = self._modules.values()
        w = bn.weight / (bn.running_var + bn.eps) ** 0.5
        w = c.weight * w[:, None, None, None]
        b = bn.bias - bn.running_mean * bn.weight / \
            (bn.running_var + bn.eps) ** 0.5
        m = torch.nn.Conv2d(w.size(1) * self.c.groups, w.size(
            0), w.shape[2:], stride=self.c.stride, padding=self.c.padding, dilation=self.c.dilation,
                            groups=self.c.groups)
        m.weight.data.copy_(w)
        m.bias.data.copy_(b)
        return m


class GRN(nn.Module):
    """ GRN (Global Response Normalization) layer
    """

    def __init__(self, dim):
        super().__init__()
        self.gamma = nn.Parameter(torch.zeros(1, 1, 1, dim))
        self.beta = nn.Parameter(torch.zeros(1, 1, 1, dim))

    def forward(self, x):
        Gx = torch.norm(x, p=2, dim=(1, 2), keepdim=True)
        Nx = Gx / (Gx.mean(dim=-1, keepdim=True) + 1e-6)
        return self.gamma * (x * Nx) + self.beta + x


class BN_Linear(torch.nn.Sequential):
    def __init__(self, a, b, bias=True, std=0.02):
        super().__init__()
        self.add_module('bn', torch.nn.BatchNorm1d(a))
        self.add_module('l', torch.nn.Linear(a, b, bias=bias))
        trunc_normal_(self.l.weight, std=std)
        if bias:
            torch.nn.init.constant_(self.l.bias, 0)

    @torch.no_grad()
    def fuse(self):
        bn, l = self._modules.values()
        w = bn.weight / (bn.running_var + bn.eps) ** 0.5
        b = bn.bias - self.bn.running_mean * \
            self.bn.weight / (bn.running_var + bn.eps) ** 0.5
        w = l.weight * w[None, :]
        if l.bias is None:
            b = b @ self.l.weight.T
        else:
            b = (l.weight @ b[:, None]).view(-1) + self.l.bias
        m = torch.nn.Linear(w.size(1), w.size(0))
        m.weight.data.copy_(w)
        m.bias.data.copy_(b)
        return m


class Block(nn.Module):

    def __init__(self, dim, drop_path=0.):
        super().__init__()
        self.dwconv = nn.Conv2d(dim, dim, kernel_size=7, padding=3, groups=dim)  # depthwise conv
        self.norm = nn.LayerNorm(dim, eps=1e-6)
        self.pwconv1 = nn.Linear(dim, 4 * dim)  # pointwise/1x1 convs, implemented with linear layers
        self.act = nn.GELU()
        self.grn = GRN(4 * dim)
        self.pwconv2 = nn.Linear(4 * dim, dim)
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()

    def forward(self, x):
        input = x
        x = self.dwconv(x)
        x = x.permute(0, 2, 3, 1)  # (N, C, H, W) -> (N, H, W, C)
        x = self.norm(x)
        x = self.pwconv1(x)
        x = self.act(x)
        x = self.grn(x)
        x = self.pwconv2(x)
        x = x.permute(0, 3, 1, 2)  # (N, H, W, C) -> (N, C, H, W)

        x = input + self.drop_path(x)
        return x


class BioBlock(torch.nn.Module):

    def __init__(self, type,
                 ed,
                 k_size,
                 drop_p
                 ):
        super().__init__()

        self.dw0 = Residual(Conv2d_BN(ed, ed, 3, 1, 1, groups=ed, bn_weight_init=0))
        self.ffn0 = Residual(FFN(ed, int(ed * 2), k_size=k_size, drop_p=drop_p))

        if type == 's':
            self.mixer = Residual(simam_module(channels=int(ed * 2)))

        self.dw1 = Residual(Conv2d_BN(ed, ed, 3, 1, 1, groups=ed, bn_weight_init=0))
        self.ffn1 = Residual(FFN(ed, int(ed * 2), k_size=k_size, drop_p=drop_p))

    def forward(self, x):
        return self.ffn1(self.dw1(self.mixer(self.ffn0(self.dw0(x)))))


class Residual(torch.nn.Module):
    def __init__(self, m, drop=0.):
        super().__init__()
        self.m = m
        self.drop = drop

    def forward(self, x):
        if self.training and self.drop > 0:
            return x + self.m(x) * torch.rand(x.size(0), 1, 1, 1,
                                              device=x.device).ge_(self.drop).div(1 - self.drop).detach()
        else:
            return x + self.m(x)


class FFN(torch.nn.Module):
    def __init__(self, endim, hidden_dim, k_size, drop_p=0.5):
        super().__init__()
        self.pw1 = Conv2d_BN(endim, hidden_dim, ks=k_size, pad=(k_size - 1) // 2)
        self.act = torch.nn.ReLU()
        self.pw2 = Conv2d_BN(hidden_dim, endim, ks=k_size, pad=(k_size - 1) // 2, bn_weight_init=0)
        # new_added
        self.drop = DropPath(drop_prob=drop_p)

    def forward(self, x):
        x = self.drop(self.pw2(self.act(self.pw1(x))))
        return x



class FeedForward(nn.Module):
    def __init__(self, dim, hidden_dim, dropout=0.):
        super().__init__()
        self.net = nn.Sequential(
            nn.Linear(dim, hidden_dim),
            nn.GELU(),
            DropPath(drop_prob=dropout),
            nn.Linear(hidden_dim, dim),
            DropPath(drop_prob=dropout)
        )

    def forward(self, x):
        x = self.net(x)
        return x


class MixerBlock(nn.Module):
    def __init__(self, dim, num_patch, token_dim, channel_dim, dropout=0.):
        super().__init__()
        self.token_mixer = nn.Sequential(
            nn.LayerNorm(dim),
            Rearrange('b n d -> b d n'),
            FeedForward(num_patch, token_dim, dropout),
            Rearrange('b d n -> b n d')

        )
        self.channel_mixer = nn.Sequential(
            nn.LayerNorm(dim),
            FeedForward(dim, channel_dim, dropout)
        )

    def forward(self, x):
        x = x + self.token_mixer(x)
        x = x + self.channel_mixer(x)
        return x


class BioUMixer(nn.Module):
    def __init__(self, in_channels, dim, num_classes, patch_size, image_size, depth, token_dim, channel_dim,
                 dropout=0.):
        super().__init__()
        assert image_size % patch_size == 0
        self.num_patches = (image_size // patch_size) ** 2  # （224/16）**2=196

        self.hid_dim = in_channels * patch_size

        self.patch_embed = nn.Sequential(Conv2d_BN(in_channels, dim // 8, 3, 2, 1),
                                         nn.LeakyReLU(),
                                         Conv2d_BN(dim // 8, dim // 4, 3, 2, 1,
                                                   ), torch.nn.ReLU(),
                                         Conv2d_BN(dim // 4, dim // 2, 3, 2, 1,
                                                   ), torch.nn.ReLU(),
                                         Conv2d_BN(dim // 2, dim, 3, 2, 1,
                                                   ),

                                         )
        self.conv_block1 = Block(dim=dim)
        self.bio_block1 = BioBlock(type='s', ed=dim // 2, k_size=7, drop_p=0.1)
        self.att1 = Residual(simam_module(channels=dim // 4))

        self.dwon1 = nn.Sequential(
            nn.Conv2d(in_channels=dim, out_channels=dim // 2, kernel_size=3, stride=1, padding=1),
            nn.MaxPool2d(kernel_size=2)
        )
        self.dwon2 = nn.Sequential(
            nn.Conv2d(in_channels=dim // 2, out_channels=dim // 4, kernel_size=3, stride=1, padding=1),
            nn.MaxPool2d(kernel_size=2)
        )
        self.dwon3 = nn.Sequential(
            nn.Conv2d(in_channels=dim // 4, out_channels=dim // 8, kernel_size=3, stride=1, padding=1),
            nn.MaxPool2d(kernel_size=2)
        )

        self.deconv1 = deconv2d_bn(in_channels=dim // 8, out_channels=dim // 4)
        self.deconv2 = deconv2d_bn(in_channels=dim // 4, out_channels=dim // 2)
        self.deconv3 = deconv2d_bn(in_channels=dim // 2, out_channels=dim)
        self.rearrange = nn.Sequential(Rearrange('b c h w -> b (h w) c'))
        self.mixer_blocks = nn.ModuleList([])
        for _ in range(depth):
            self.mixer_blocks.append(MixerBlock(dim, self.num_patches, token_dim, channel_dim, dropout))

        #
        self.layer_normal = nn.LayerNorm(dim)

        #
        self.mlp_head = nn.Sequential(
            nn.GELU(),
            BN_Linear(dim, dim // 2),
            # DropPath(0.3),

            nn.GELU(),
            BN_Linear(dim // 2, dim // 4),
            # DropPath(0.2),

            nn.GELU(),
            BN_Linear(dim // 4, num_classes),
            # DropPath(0.1)
        )

    def forward(self, x):
        # x = self.to_embedding(x)
        x = self.patch_embed(x)
        conv1 = self.conv_block1(x)
        dwon1 = self.dwon1(conv1)
        bio1 = self.bio_block1(dwon1)
        dwon2 = self.dwon2(bio1)
        att1 = self.att1(dwon2)
        dwon3 = self.dwon3(att1)

        de1 = self.deconv1(dwon3)
        plus1 = att1 + de1
        att2 = self.att1(plus1)
        de2 = self.deconv2(att2)
        plus2 = de2 + bio1
        bio2 = self.bio_block1(plus2)
        de3 = self.deconv3(bio2)
        plus3 = conv1 + de3
        x = self.conv_block1(plus3)
        x = self.rearrange(x)
        for mixer_block in self.mixer_blocks:
            x = mixer_block(x)
        x = self.layer_normal(x)
        encoding = x.mean(dim=1)

        x = self.mlp_head(encoding)
        return x, encoding


